// Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

import "ndarray.proto";

package cogment_lab.spaces;

message Discrete {
  int32 n = 1;
  int32 start = 2;
}

message Box {
  nd_array.Array low = 2;
  nd_array.Array high = 3;
}

message MultiBinary {
  nd_array.Array n = 1;
}

message MultiDiscrete {
  nd_array.Array nvec = 1;
}

message Dict {
  message SubSpace {
    string key = 1;
    Space space = 2;
  }
  repeated SubSpace spaces = 1;
}

message Space {
  oneof kind {
    Discrete discrete = 1;
    Box box = 2;
    Dict dict = 3;
    MultiBinary multi_binary = 4;
    MultiDiscrete multi_discrete = 5;
  }
}
