// Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

package cogment_lab.nd_array;

enum DType {
  DTYPE_UNKNOWN = 0;
  DTYPE_FLOAT32 = 1;
  DTYPE_FLOAT64 = 2;
  DTYPE_INT8 = 3;
  DTYPE_INT32 = 4;
  DTYPE_INT64 = 5;
  DTYPE_UINT8 = 6;
}

message Array {
  DType dtype = 1;
  repeated uint32 shape = 2;
  bytes raw_data = 3;
  bytes npy_data = 4;
  repeated double double_data = 5;
  repeated sint32 int32_data = 6;
  repeated sint64 int64_data = 7;
  repeated uint32 uint32_data = 8;
}
